********************************************************************
*** Name.....: CURSORMGR.PRG
*** Author...: Andy Kramek & Marcia Akins
*** Date.....: 10/15/2008
*** Notice...: Copyright (c) 2008 Tightline Computers, Inc
*** Compiler.: Visual FoxPro 09.00.0000.3504 for Windows 
*** Function.: Class Definition for a cursor manager object
*** Returns..: Object Reference to the cursor manager or NULL
********************************************************************
*** SelfTest Options:
********************************************************************
*** 1 = Check the Table Access Method
*** 2 = Generate an Error to test logging
*** 3 = Call the SelectCursor Method with no key - should cause Error
*** 4 = Call the SelectCursor Method with an invalid key - should cause Error
*** 5 = Call the SelectCursor Method with a Create Cursor definition
*** 6 = Call the SelectCursor Method with a SQL Query definition
*** 7 = Call the SelectCursor Method with an Open Table definition

********************************************************************
#DEFINE CRLF CHR(13) + CHR(10) 

********************************************************************
*** Code used for command line testing with "DO cursormgr":
********************************************************************
RELEASE oT
PUBLIC oT
oT = CREATEOBJECT( 'cursormgr' )
********************************************************************

********************************************************************
*** Class definition starts here
********************************************************************
DEFINE CLASS cursormgr AS SESSION

  *** We need a collection for error messages
  PROTECTED oErrObj
  oErrObj = NULL
  
  *** And a property for the name of the definition table to use
  *** This is so that we can use an access method to ensure it is open
  DefTableName = 'cursordef'
  
  *** And the name of the index on the metadata table
  DefIndex = 'cursor_nme'  

  *** And a property to preserve the object's own DSID when changing to 
  *** Create or Open a cursor
  PROTECTED SelfDSID 
  SelfDSID = 0
  
  ********************************************************************
  *** [E] SELECTCURSOR: Open or Create the specified entity based on its key name
  ********************************************************************
  FUNCTION SelectCursor ( tcKeyName, tnDataSessionID )
  LOCAL llRetVal, lcType, loCurDef, lcAlias, lcToCall
    WITH This
      TRY
        *** First we must save off our own datasession - in case the DataSessionID gets changed
        *** This will be restored in the FINALLY clause of the TRY...CATCH so it always resets
        .SelfDSID = .DataSessionID

        *** Did we get something that looks like a key name?
        IF .ValidateParam( tcKeyName, 'C' ) < 1
          ERROR 'A Key Name must be passed when calling SelectCursor'
        ELSE
          IF NOT SEEK( UPPER( ALLTRIM( tcKeyName )), .DefTableName, .DefIndex )
            ERROR 'Key ' + tcKeyName + ' is not a valid identifier in ' + .DefTableName
          ENDIF
        ENDIF
        
        *** If we get to here we actually do have a cursor definition record for this name
        *** We will already be in the correct work area because the Access Method sets it
        *** So all we need to do is to get all the data into an object
        SCATTER NAME loCurDef MEMO

        *** If we didn't get a datasessionid, assume it is "1" - but either way change now
        lnDSID = IIF( .ValidateParam( tnDataSessionID, 'N' ) = 1, tnDataSessionid, 1 )
        SET DATASESSION TO ( lnDSID )

        *** Get the alias and check to see if it already exists
        lcAlias = UPPER(ALLTRIM( loCurDef.alias_nme ))
        IF USED( lcAlias )
          *** Just select it and get out of here
          SELECT (lcAlias)
        ELSE
          *** Call the correct creation method based on the type
          llRetVal = EVALUATE( "This.OpenCursor" + loCurDef.cursor_typ + "( loCurDef )" )
        ENDIF
   
        *** If we get here, all is well
        llRetVal = .T.

      CATCH TO loErr
        This.LogError( loErr )
        llRetVal = .F.
      FINALLY
        *** Always restore the datasession
        SET DATASESSION TO (This.SelfDSID)
      ENDTRY

    ENDWITH
    *** Return Final Status    
    RETURN llRetVal
  ENDFUNC && SelectCursor 

  ********************************************************************
  *** [E] SELFTEST: Run a test operation
  ********************************************************************
  FUNCTION SelfTest( tnTest )
  LOCAL lcTable, lcResult, loErr, lcMsg
    WITH This
      TRY
        *** Default the response
        lcResult = 'Success!'
        DO CASE
          CASE tnTest = 1
            *** Check the table access method
            *** This should select the specified alias
            lcTable = .DefTableName  
            lcResult = 'Successfully Selected Alias = ' + ALIAS()

          CASE tnTest = 2
            *** Check the error logging method
            ERROR 'Testing the Error Logging'

          CASE tnTest = 3
            *** Call the SelectCursor Method with no key - should cause Error
            .SelectCursor()
          
          CASE tnTest = 4 
            *** Call the SelectCursor Method with an invalid key - should cause Error
            .SelectCursor( 'nonsuch' )
          
          CASE tnTest = 5
            *** Create a cursor using the  AKTest/CRE definition
            .SelectCursor( 'CMCreTest' ) 
                
          CASE tnTest = 6
            *** Create a cursor by running a SQL Select Statement
            .SelectCursor( 'CMSQLTest' )
            
          CASE tnTest = 7
            *** Open the specified table  
            .SelectCursor( 'CMTblTest' )
            
          OTHERWISE
            *** Do Nothing at all  
        ENDCASE
   
      CATCH TO loErr
        This.LogError( loErr )

      ENDTRY
    ENDWITH
    *** Show the result
    loErr = This.GetErrors()
    IF loErr.nErrors > 0
      loErr.ShowErrors
    ELSE
      lcMsg = 'Selftest Option ' + TRANSFORM( tnTest ) + ' reported: ' + CRLF + lcResult
      MESSAGEBOX( lcMsg, 0, 'Selftest Report' )
    ENDIF
  ENDFUNC && SelfTest

  ********************************************************************
  *** [E] GETERRORS(): Returns the contents of the error collection upon demand
  ***                  in a parameter object and clears the error collection
  ********************************************************************
  FUNCTION GetErrors() AS String
    LOCAL loParamObj, lnErrCnt, lnCnt, loErr, lcErrs, llSaveErrorLog
    *** Create the parameter object
    loParamObj = CREATEOBJECT( 'xErrorObj' )
    IF This.oErrObj.Count = 0
      *** No errors to display, return an empty string
      lcErrs = 'No Errors Found'
      loParamObj.nerrors = 0
    ELSE
      *** Populate the Parameter Object here
      WITH loParamObj
        *** At least one error, so add Counter and Array
        lnErrCnt = This.oErrObj.Count 
        .nerrors = lnErrCnt 
        *** Set the error string and counters
        lnCnt = 1
        FOR EACH loErr IN This.oErrObj
          *** Add a row to the error array
          DIMENSION .aErrors[lnCnt, 5]
          *** Log the error as Character Strings
          .aErrors[lnCnt, 1] = TRANSFORM( loErr.errorno )  && Error Number
          .aErrors[lnCnt, 2] = TRANSFORM( loErr.message )  && Error Message
          .aErrors[lnCnt, 3] = TRANSFORM( loErr.procedure )  && Prog/Method (if any)
          .aErrors[lnCnt, 4] = TRANSFORM( loErr.details )  && Error Details (if any)
          .aErrors[lnCnt, 5] = TRANSFORM( loErr.linecontents )  && Error Line content (if any)
          *** Increment the counter
          lnCnt = lnCnt + 1 
        NEXT
      ENDWITH
      
      *** We always want to clear the error collection when this is called
      This.oErrObj.Remove(-1)
    ENDIF

    *** Return Errors
    RETURN loParamObj
  ENDFUNC
  *** eof GetErrors()
  
  ********************************************************************
  *** [E] INIT(): Standard Initialization Method
  ********************************************************************
  FUNCTION INIT()
    WITH This
      *** Create the Error Collection object
      .oErrObj = CREATEOBJECT( 'collection' )
    ENDWITH
  ENDFUNC
  *** eof Init()
  
  ********************************************************************
  *** [P] OPENCURSORSQL: Run the SQL Query define in the memo
  ********************************************************************
  PROTECTED FUNCTION OpenCursorSQL( toCurDef )
  LOCAL llRetVal
    WITH This
      TRY
        *** Get the cursor specification
        lcCursor = UPPER( ALLTRIM( toCurDef.alias_nme )) 
        lcCmd = ALLTRIM( CHRTRAN( toCurDef.cursor_def, CHR(13)+CHR(10), '' ))
        lcCmd = lcCmd + " INTO CURSOR " + lcCursor
        *** Execute it 
        &lcCmd
        *** And Select it
        SELECT (lcCursor)
        *** If we get here, all is well
        llRetVal = .T.

      CATCH TO loErr
        This.LogError( loErr )
        llRetVal = .F.
      ENDTRY
    ENDWITH
    *** Return Final Status    
    RETURN llRetVal
  ENDFUNC && OpenCursorSQL
  
  ********************************************************************
  *** [P] OPENCURSORCRE: Create the cursor as defined in the memo
  ********************************************************************
  PROTECTED FUNCTION OpenCursorCRE( toCurDef )
  LOCAL llRetVal, lcCmd, lcCursor
    WITH This
      TRY
        *** Get the cursor specification
        lcCursor = UPPER( ALLTRIM( toCurDef.alias_nme )) 
        lcCmd = "CREATE CURSOR " + lcCursor + "( "
        lcCmd = lcCmd + ALLTRIM( CHRTRAN( toCurDef.cursor_def, CHR(13)+CHR(10), '' )) + ")"
        *** Execute it 
        &lcCmd
        *** And Select it
        SELECT (lcCursor)
        *** If we get here, all is well
        llRetVal = .T.
      CATCH TO loErr
        This.LogError( loErr )
        llRetVal = .F.
      ENDTRY
    ENDWITH
    *** Return Final Status    
    RETURN llRetVal
  ENDFUNC && OpenCursorCRE
  
  ********************************************************************
  *** [P] OPENCURSORTBL: Open a Table with the specified alias
  ********************************************************************
  PROTECTED FUNCTION OpenCursorTbl( toCurDef )
  LOCAL llRetVal, lcAlias, lcCmd
    WITH This
      TRY
        *** Get the cursor specification
        lcAlias = UPPER( ALLTRIM( toCurDef.alias_nme )) 
        lcTable = FORCEEXT( ALLTRIM( CHRTRAN( toCurDef.cursor_def, CHR(13)+CHR(10), '' )), 'DBF' )
        *** Do we have the table name?
        IF NOT FILE( lcTable )
          ERROR lcTable + " is not in the current search path"
        ENDIF
        
        *** If we get to here, open the table with the specified Alias
        *** The parent method already checked for the Alias name in use
        lcCmd = "USE " + lcTable + " AGAIN IN 0 ALIAS " + lcAlias
        *** Execute it 
        &lcCmd
        *** And Select it
        SELECT (lcAlias)
        *** If we get here, all is well
        llRetVal = .T.
   
      CATCH TO loErr
        This.LogError( loErr )
        llRetVal = .F.
      ENDTRY
    ENDWITH
    *** Return Final Status    
    RETURN llRetVal
  ENDFUNC && OpenCursorTbl

  ********************************************************************
  *** [P] VALIDATEPARAM: Check Parameter is of expected type
  *** Returns........:  1 if Value is correct data type and not empty
  *** ...............:  0 if Value is correct data type, but EMPTY()
  *** ...............: -1 if Value is wrong data type
  ********************************************************************
  PROTECTED FUNCTION ValidateParam( tuTestParam AS Variant, tcReqType AS String )
  LOCAL lnRetVal, lcType
    ***  Check for Type first (Type may be a set of data types, e.g. "DT"
    IF VARTYPE( tuTestParam ) $ UPPER( tcReqType )
      IF tcReqType # "O"
        ***  Data Type is OK, but is it EMPTY()
        lnRetVal = IIF( EMPTY( tuTestParam ), 0, 1 )
      ELSE
        *** Check for NULL object reference
        lnRetVal = IIF( ISNULL( tuTestParam ), 0, 1 )
      ENDIF && tcReqType # "O"
    ELSE && VARTYPE( tuTestParam ) $ UPPER( tcReqType )
      ***  Wrong Data Type!
      lnRetVal = -1
    ENDIF && VARTYPE( tuTestParam ) $ UPPER( tcReqType )
    ***  Return Result
    RETURN lnRetVal
  ENDFUNC
  *** eof: ValidateParam()  
  
  ********************************************************************
  *** [P] DEFTABLENAME_ACCESS: Access method for the Cursor Definition Table property
  ********************************************************************
  PROTECTED FUNCTION DefTableName_Access()
  LOCAL lcRetVal, lcAlias, lcTable
    WITH This
      TRY
        lcAlias = JUSTSTEM( .DefTableName ) 
        lcTable = FORCEEXT( lcAlias, 'dbf' )
        IF NOT USED( lcAlias )
          USE ( lcTable ) IN 0 AGAIN SHARED ALIAS ( lcAlias )
        ENDIF

        *** And Select it
        SELECT (lcAlias)
        
        *** If we get here, all is well
        lcRetVal = ALIAS()

      CATCH TO loErr
        This.LogError( loErr )
        lcRetVal = ''
      ENDTRY
    ENDWITH
    *** Return Final Status    
    RETURN lcRetVal
  ENDFUNC && DefTableName_Access
  
  *******************************************************************
  *** ERROR(): Override the Native VFP Error Method to redirect 
  ***          runtime errors to the custom LogError() Method
  *******************************************************************
  FUNCTION ERROR( tnErrNum, tcErrProc, tnErrLine )
    *** Log the error - passing the Error text instead of the Line Number
    This.LogError( tnErrNum, MESSAGE(), tcErrProc )
    RETURN
  ENDFUNC
  *** eof Error()

  *******************************************************************
  *** [P] LOGERROR(): Standard Log Error Method
  *** Called from the Error method and from CATCH block as needed
  *** Adds an error to the internal errors collection
  *******************************************************************
  PROTECTED FUNCTION LOGERROR( tuErrNum, tcErrMsg, tcMethod )
    LOCAL lcType, loErr
    
    *** Did we get a VFP Error, or an Exception Object?
    lcType = VARTYPE( tuErrNum )
    IF lcType # 'O'
      *** This is not an exception object, so we need to manufacture one
      loErr = CREATEOBJECT( 'exception' )
      WITH loErr
          .errorno      = tuErrNum
          .procedure    = IIF( EMPTY( tcMethod ), "", ALLTRIM( tcMethod ) )
          .message      = IIF( EMPTY( tcErrMsg ), "", ALLTRIM( tcErrMsg ) )
      ENDWITH
    ELSE
      *** This is an exception object already, just assign it
      loErr = tuErrNum
    ENDIF

    *** Now whatever the source we have an exception object so add it to the collection
    .oErrObj.Add( loErr )

    *** If we are running interactively, show the error immediately
    IF _VFP.StartMode = 0 
      *** Display the error (interactive Mode only)
      MESSAGEBOX( loErr.message + CRLF + "Use GetErrors() to retrieve all details", 16, 'Error!' )
    ENDIF

    *** Just Return Success!
    RETURN .T.
  ENDFUNC
  *** eof: LogError()

ENDDEFINE

********************************************************************
*** Parameter Object Definition, used by GETERRORS() to pass error 
*** information back from the Manager object
********************************************************************
DEFINE CLASS xErrorObj AS custom

  *** Standard Properties 
  nErrors = 0
  DIMENSION aErrors[1]

  ********************************************************************
  *** [E] SHOWERRORS: Displays the contents of the error collection
  ********************************************************************
  FUNCTION ShowErrors()
  LOCAL lcErrs, lcText, lnSel, lcFile, lnCnt
    WITH This
      IF .nErrors > 0
        lcErrs = ''
        *** Extract the error information
        FOR lnCnt = 1 TO .nErrors
          *** Build the error string for display
          lcErrs = lcErrs + IIF( NOT EMPTY( lcErrs ), CRLF, "" ) + "Error " + TRANSFORM( .aErrors[lnCnt, 1] )
          lcErrs = lcErrs + " ( " + ALLTRIM( .aErrors[lnCnt, 2] ) + " )" + CRLF
          lcErrs = lcErrs + "In : " + ALLTRIM( .aErrors[lnCnt, 3] )
          IF NOT EMPTY( .aErrors[lnCnt, 4] )
            lcErrs = lcErrs + CRLF + "Details : " + ALLTRIM( .aErrors[lnCnt, 4] ) 
          ENDIF
          IF NOT EMPTY( .aErrors[lnCnt, 5] )
            lcErrs = lcErrs + CRLF + "Line : " + ALLTRIM( .aErrors[lnCnt, 5] )
          ENDIF
          
        NEXT
      ELSE
        lcErrs = "No Errors Recorded"
      ENDIF
      *** Add on the display information
      lcText = lcErrs + CRLF + "==============================" + CRLF
      lcText = lcText + "Save this list to a text file?" + CRLF
      *** Get the details
      lnSel = MESSAGEBOX( lcText, 36, "Error Report" )
      IF lnSel = 6 && The user said YES!
        lcFile = 'ErrorLog' 
        lcFile = PUTFILE( "Name As", lcFile, 'txt' )
        IF NOT EMPTY( lcFile )
          STRTOFILE( lcErrs, lcFile, 0 )
        ENDIF
      ENDIF
    ENDWITH
  ENDFUNC
  *** eof ShowErrors()
  
ENDDEFINE